package Data::Convertor::Mongo;
use strict;
use warnings;

#*******************************************************************************
# DOCUMENTATION SECTION
#*******************************************************************************

=head1 NAME

Data::Convertor::Mongo - Convert values od various types into MongoDB optimal representation

=head1 SYNOPSIS

 use Data::Convertor::Mongo;

=head1 DESCRIPTION


=head1 AUTHOR

 Jan Mach
 Cesnet, z.s.p.o
 jan.mach@cesnet.cz
 http://www.cesnet.cz

=head1 COPYRIGHT

This program is free software; you can redistribute
it and/or modify it under the same terms as Perl itself.

The full text of the license can be found in the
LICENSE file included with this module.

=head1 SEE ALSO

perl(1), Data::Convertor(3).

=head1 FUNCTION REFERENCE

=over 4

=cut

#*******************************************************************************
# LIBRARY LOADING SECTION
#*******************************************************************************

#-- Perl core modules ---------------------------------------------------------#
use Carp;
use Scalar::Util 'blessed';

#use Data::Dumper; #-+-> DEVEL ONLY <-+-#
#use Smart::Comments; #-+-> DEVEL ONLY <-+-#

#-- Perl CPAN modules ---------------------------------------------------------#

#-- Custom application modules ------------------------------------------------#
use Data::Convertor;

# Modules from Cesnet Toolkit
use Log::Loger;

#*******************************************************************************
# CONSTANTS AND GLOBAL VARIABLES DEFINITION SECTION
#*******************************************************************************

#-- Constants -----------------------------------------------------------------#

#-- Static public class variables (our) ---------------------------------------#

#-- Static protected class variables (my) -------------------------------------#

#*******************************************************************************
# INITIALIZATION AND CLEANUP SECTION
#*******************************************************************************

#-- Module initializations ----------------------------------------------------#
BEGIN {
    use vars qw($VERSION $DEVEL @ISA);
    $VERSION = '0.1';
    $DEVEL   = 0;
    @ISA     = ('Data::Convertor');
}


#-- Module clean-up code (global destructor) ----------------------------------#
END {

}

#*******************************************************************************
# IMPLEMENTATION SECTION
#*******************************************************************************

=item convert_value($$$) [PUBLIC]

  Usage     : my $result = $instance->convert_value($mpath, $value, $operator)
  Purpose   : Convert given value to the datatype expected for node on given MPath
  Arguments : STRING $mpath - path to value
              STRING $value - value as string
              STRING $operator - operator as string
  Returns   : Value in hash and optimal representation (INT, STRING, ...)
  Throws    : Croaks, if not invoked on object
  Comments  : Some conversions require
  See Also  :

=cut

sub convert_value($$$)
{
    my $self = shift;
    confess "Instance method invoked on class" unless blessed($self);
    my ($dpath, $value, $operator) = @_;
    confess "DPath must be given as argument" unless defined($dpath);
    confess "Value must be given as argument" unless defined($value);

    my ($conversion, $type, $cmin, $cmax, $sub, $ipv);

    # Prepare the DPath
    $conversion->{DPATH} = $dpath;
    # We must translate '/' in DPath to '.', which is the separator in MongoDB queries
    $conversion->{DPATH} =~ s|/|.|g;
    # And we must remove the index specifications
    $conversion->{DPATH} =~ s/(\[[\d]+\])//g;

    # Prepare the rest of the arguments
    $conversion->{PATH}     = $conversion->{DPATH};
    $conversion->{VALUE}    = $value;
    $conversion->{OPERATOR} = $operator;
    $conversion->{REPLACE}  = 0;

    if ($conversion->{DPATH} eq "ts") {
        ($type, $cmin) = $self->_convert_with_conversion('datetime_to_nbi', $conversion->{VALUE});
        $conversion->{VALUE} = $cmin;
    }

    #Log::Loger->info(__PACKAGE__, "Conversion: ", Dumper ($conversion) ) if $DEVEL;
    #Log::Loger->info(__PACKAGE__, "PATH: ", $conversion->{PATH} ) if $DEVEL;
    #Log::Loger->info(__PACKAGE__, "VALUE: ", $conversion->{VALUE} ) if $DEVEL;

    return $conversion;
}

#-------------------------------------------------------------------------------
# Private interface
#-------------------------------------------------------------------------------

# _init [PROTECTED]
#
#  Usage     : return $self->_init(@_);
#  Purpose   : Initialize the new Data::Convertor instance
#  Arguments :
#  Returns   : Data::Convertor reference
#  Throws    : Croaks, if invoked on class
#  Comments  : This method must never be called directly, it is supposed to be overloaded by subclasses.
#  See Also  :

sub _init
{
    my $self = shift;
    confess "Instance method invoked on class" unless blessed($self);

    return $self;
}

=pod

=back

=cut

1;
