package Mentat::Stats::Data;
use strict;
use warnings;

#*******************************************************************************
# DOCUMENTATION SECTION
#*******************************************************************************

=head1 NAME

Mentat::Stats::Data - Short module description

=head1 SYNOPSIS

 use Mentat::Stats::Data;

=head1 DESCRIPTION

Extensive module description

=head1 AUTHOR

 Jan Mach
 Cesnet, z.s.p.o
 jan.mach@cesnet.cz
 http://www.cesnet.cz

=head1 COPYRIGHT

This program is free software; you can redistribute
it and/or modify it under the same terms as Perl itself.

The full text of the license can be found in the
LICENSE file included with this module.

=head1 SEE ALSO

perl(1), Mentat::Stats(3).

=head1 FUNCTION REFERENCE

=over 4

=cut

#*******************************************************************************
# LIBRARY LOADING SECTION
#*******************************************************************************

#-- Perl core modules ---------------------------------------------------------#
use Carp;
use Scalar::Util 'blessed';
use Storable 'dclone';

use Number::Format qw(:subs :vars);

use Data::Dumper; #-+-> DEVEL ONLY <-+-#
#use Smart::Comments; #-+-> DEVEL ONLY <-+-#

#-- Perl CPAN modules ---------------------------------------------------------#

#-- Custom application modules ------------------------------------------------#
use Color::Palette;
use Chart::Palette;
use Mentat::Stats;
use Mentat::Stats::Charter;

#*******************************************************************************
# CONSTANTS AND GLOBAL VARIABLES DEFINITION SECTION
#*******************************************************************************

#-- Constants -----------------------------------------------------------------#
use constant IP_THRESHOLD  => 200;
use constant ID_THRESHOLD  => 200;
use constant TOP_THRESHOLD => 20;

use constant ST_INTERNAL => 'internal';
use constant ST_EXTERNAL => 'external';
use constant ST_OVERALL  => 'overall';

#-- Static public class variables (our) ---------------------------------------#

#-- Static protected class variables (my) -------------------------------------#

#*******************************************************************************
# INITIALIZATION AND CLEANUP SECTION
#*******************************************************************************

#-- Module initializations ----------------------------------------------------#
BEGIN {
    use vars qw($VERSION $DEVEL @ISA);
    $VERSION = '0.1';
    $DEVEL   = 0;
    @ISA     = ('Mentat::Stats');
}


#-- Module clean-up code (global destructor) ----------------------------------#
END {

}

#*******************************************************************************
# IMPLEMENTATION SECTION
#*******************************************************************************

=item DESTROY() [PUBLIC]

  Usage     : This method must not be used directly
  Purpose   : Class destructor
  Arguments : None
  Returns   : Nothing
  Throws    : Dies, if not invoked on instance

=cut

sub DESTROY()
{
    my $self = shift;
    confess "Instance method not invoked on object instance" unless blessed($self);
    ### Destroying Mentat::Stats::Data instance...
}

sub _escape($)   { my $v = shift; $v =~ s/\./(dot)/g;  return $v; }
sub _unescape($) { my $v = shift; $v =~ s/\(dot\)/./g; return $v; }
sub _unescape_data($)
{
    my $d = shift;
    my $rv = {};
    my $v;
    for my $k (keys %$d) {
        $v = $d->{$k};
        $rv->{_unescape($k)} = $v;
    }
    return $rv;
}

#-------------------------------------------------------------------------------

=item evaluate_alerts($;$$) [PUBLIC]

  Usage     : my $stats = $reportstats->evaluate_alerts($messages, $stats);
  Purpose   : Evaluate statistics for list of alerts
  Arguments : ARRAY REFERENCE $messages - list of messages [MANDATORY]
  Returns   : HASH REFERENCE $stats
  Throws    : Dies, if not invoked on instance

=cut

sub evaluate_alerts($;$$)
{
    my $self = shift;
    confess "Instance method not invoked on object instance" unless blessed($self);
    my ($messages, $stats, $brief) = @_;

    $messages = [$messages] unless ref $messages eq 'ARRAY';
    $stats    = {}          unless $stats and ref $stats eq 'HASH';

    my ($vals, $val_id, $tmp, $dn, $sn, $key);
    foreach my $msg (@$messages) {
        # Save the message identifier
        $val_id = $msg->id();
        push(@{$stats->{'list_ids'}}, $val_id);

        # Calculate other statistics
        foreach my $rule (
            ['ips',         'Source/IP4'],
            #['ip4s',        'Source/IP4'],
            #['ip6s',        'Source/IP6'],
            ['analyzers',   'Node[#]/SW'],
            ['categories',  'Category'],
            ['detectors',   'Node[#]/Name'],
        ) {
            $vals = $msg->path_values($rule->[1]);
            if (@$vals) {
                foreach my $v (@$vals) {
                    $v =~ s/\./(dot)/g;
                    $stats->{$rule->[0]}->{$v}++; # In MongoDB database, hash keys must not contain dots '.'
                }
                if ($rule->[0] eq 'categories') {
                    $tmp = join('/', sort map { $_ =~ s/\./(dot)/g; $_; } @$vals);
                    $stats->{'category_sets'}->{$tmp}++;
                }
            }
            else {
                $stats->{$rule->[0]}->{'_unknown'}++;
            }
        }

        # Calculate statistics for last analyzer/detector concatenation
        $dn = $msg->path_values('Node[#]/Name');
        $sn = $msg->path_values('Node[#]/SW');
        for my $va (@$dn) {
            if (@$sn) {
                for my $vb (@$sn) {
                    $key = join('/', $va, $vb);
                    $key =~ s/\./(dot)/g; # In MongoDB database, hash keys must not contain dots '.'
                    $stats->{'detectorsws'}->{$key}++;
                }
            }
            else {
                $key = "".$va;
                $key =~ s/\./(dot)/g; # In MongoDB database, hash keys must not contain dots '.'
                $stats->{'detectorsws'}->{$key}++;
            }
        }
    }

    # Calculate some total counts
    $stats->{'cnt_alerts'}        = scalar @$messages;
    $stats->{'cnt_ips'}           = scalar keys %{$stats->{'ips'}};
    $stats->{'cnt_analyzers'}     = scalar keys %{$stats->{'analyzers'}};
    $stats->{'cnt_categories'}    = scalar keys %{$stats->{'categories'}};
    $stats->{'cnt_category_sets'} = scalar keys %{$stats->{'category_sets'}};
    $stats->{'cnt_detectors'}     = scalar keys %{$stats->{'detectors'}};
    $stats->{'cnt_detectorsws'}   = scalar keys %{$stats->{'detectorsws'}};

    # Prepare various lists
    $stats->{'list_ips'}           = [sort {$stats->{'ips'}->{$b} <=> $stats->{'ips'}->{$a}} keys %{$stats->{'ips'}}];
    $stats->{'list_analyzers'}     = [sort keys %{$stats->{'analyzers'}}];
    $stats->{'list_categories'}    = [sort keys %{$stats->{'categories'}}];
    $stats->{'list_category_sets'} = [sort keys %{$stats->{'category_sets'}}];
    $stats->{'list_detectors'}     = [sort keys %{$stats->{'detectors'}}];
    $stats->{'list_detectorsws'}   = [sort keys %{$stats->{'detectorsws'}}];

    # Shorten some data on purpose
    if ($brief) {
        # Do not store event IDs at all
        delete $stats->{'list_ids'};
        # Store only limited amount of IPs (toplist with threshold)
        $stats->{'list_ips'} = [splice(@{$stats->{'list_ips'}}, 0, TOP_THRESHOLD())];
        # Get rid of the unnecessary entries (using temporary array)
        my $tmp = {};
        foreach my $i (@{$stats->{'list_ips'}}) {
            $tmp->{$i} = $stats->{'ips'}->{$i};
        }
        $stats->{'ips'} = $tmp;
    }
    # Or automaticaly to keep them brief
    else {
        # Store only limited amount of event IDs
        if ($stats->{'cnt_alerts'} > ID_THRESHOLD()) {
            $stats->{'list_ids'} = [splice(@{$stats->{'list_ids'}}, 0, ID_THRESHOLD())];
        }
        # Store only limited amount of IPs
        if ($stats->{'cnt_ips'} > IP_THRESHOLD()) {
            $stats->{'list_ips'} = [splice(@{$stats->{'list_ips'}}, 0, IP_THRESHOLD())];
            # Get rid of the unnecessary entries (using temporary array)
            my $tmp = {};
            foreach my $i (@{$stats->{'list_ips'}}) {
                $tmp->{$i} = $stats->{'ips'}->{$i};
            }
            $stats->{'ips'} = $tmp;
        }
    }

    # In MongoDB database, hash keys must not contain dots '.', perform backward conversion
    foreach my $t ('list_ips', 'list_analyzers', 'list_categories', 'list_category_sets', 'list_detectors', 'list_detectorsws') {
        foreach my $item (@{$stats->{$t}}) {
            $item =~ s/\(dot\)/./g;
        }
    }

    #print STDERR "STATS:".Dumper($stats);

    return $stats;
}

sub _chart_add_dataset($$)
{
    my $self = shift;
    confess "Instance method not invoked on object instance" unless blessed($self);
    my ($data, $name) = @_;

    my $dataset = {};

    # Perform the sorting in descending order and put the data into chart structures
    foreach my $key (sort {$data->{$b} <=> $data->{$a}} keys %$data) {
        push(@{$dataset->{'data'}->[0]}, $key);
        push(@{$dataset->{'data'}->[1]}, $data->{$key});
        $dataset->{'cnt'}++;
        $dataset->{'sum'} += $data->{$key};
        $dataset->{'min'}  = $data->{$key} if not $dataset->{'min'} or $dataset->{'min'} > $data->{$key};
        $dataset->{'max'}  = $data->{$key} if not $dataset->{'max'} or $dataset->{'max'} < $data->{$key};
    }
    $dataset->{'avg'} = $dataset->{'sum'} / $dataset->{'cnt'} if $dataset->{'cnt'};

    # Calculate dataset median
    my @values = sort { $a <=> $b } values %$data;
    my $size   = scalar @values;
    if ($size) {
        $dataset->{'vals'} = [@values];
        if ($size % 2) { $dataset->{'med'} = $values[((($size+1)/2)-1)]; }
        else           { $dataset->{'med'} = ($values[(($size/2)-1)] + $values[((($size+2)/2)-1)]) / 2; }
    }

    # Perform dataset thresholding
    $dataset->{'data'} = $self->_chart_toplist_threshold($dataset->{data});

    return $dataset;
}

#-------------------------------------------------------------------------------

=item evaluate_logs_reporting($;$) [PUBLIC]

  Usage     : my $stats = $reportstats->evaluate_logs_reporting($logs);
  Purpose   : Evaluate statistics for report processing logs
  Arguments : ARRAY REFERENCE $messages - list of processing logs [MANDATORY]
  Returns   : HASH REFERENCE $stats
  Throws    : Dies, if not invoked on instance

=cut

sub evaluate_logs_reporting($;$)
{
    my $self = shift;
    confess "Instance method not invoked on object instance" unless blessed($self);
    my ($reporting_logs, $adhoc) = @_;
    die "Reporting logs must be given as argument\n" unless $reporting_logs;

    # Storage for calculated statistics
    my $stats = {};
    return $stats unless $reporting_logs and ref $reporting_logs eq 'ARRAY' and scalar @$reporting_logs;

    my @stat_calc_keys = ('events_all', 'events_proc', 'events_blocked', 'reports_all',
                          'duration', 'duration_fetch', 'duration_aggr',
                          'speed_fetch', 'speed_aggr',
                          'dns_resolved_host', 'dns_resolved_ip', 'dns_unresolved');

    my (@keys, $size, $tmp);
    my $i = 0;
    foreach my $l (@$reporting_logs) {
        # Calculate the time period
        $stats->{ts_from} = $l->{ts_from} if not $stats->{ts_from} or $stats->{ts_from} > $l->{ts_from};
        $stats->{ts_to}   = $l->{ts_to}   if not $stats->{ts_to}   or $stats->{ts_to} < $l->{ts_to};

        # Calculate fetching and aggregation speed, if not already present
        if (not $l->{speed_fetch} and $l->{duration_fetch}) {
            $l->{speed_fetch} = $l->{events_proc} / $l->{duration_fetch};
        }
        if (not $l->{speed_aggr} and $l->{duration_aggr}) {
            $l->{speed_aggr} = $l->{events_proc} / $l->{duration_aggr};
        }

        # Calculate batch of reporting log statistics
        foreach my $key (@stat_calc_keys) {
            $stats->{$key.'_min'} = $l->{$key} if not $stats->{$key.'_min'} or $stats->{$key.'_min'} > $l->{$key};
            $stats->{$key.'_max'} = $l->{$key} if not $stats->{$key.'_max'} or $stats->{$key.'_max'} < $l->{$key};
            $stats->{$key.'_sum'} += $l->{$key};
            $stats->{$key.'_cnt'} += 1;
        }

        # Calculate email destination statistics
        if ($l->{reports_mails}) {
            @keys = keys %{$l->{reports_mails}};
            $size = scalar @keys;

            # Number of report destinations
            $stats->{reports_dest_min} = $size if not $stats->{reports_dest_min} or $stats->{reports_dest_min} > $size;
            $stats->{reports_dest_max} = $size if not $stats->{reports_dest_max} or $stats->{reports_dest_max} < $size;
            $stats->{reports_dest_sum} += $size;
            $stats->{reports_dest_cnt} += 1;

            # Statistics for each destination
            foreach my $k (@keys) {
                unless (ref $k) {
                    $stats->{reports_dests}->{$k} += $l->{reports_mails}->{$k};
                }
                else {
                    my $x = join(';', @$k);
                    $stats->{reports_dests}->{$x} += $l->{reports_mails}->{$x};
                }
            }
        }

        # Calculate DNS effectivity
        if ($l->{dns_resolved_ip} + $l->{dns_resolved_host} + $l->{dns_unresolved}) {
            $l->{dns_effect} = (($l->{dns_resolved_ip} + $l->{dns_resolved_host})/($l->{dns_resolved_ip} + $l->{dns_resolved_host} + $l->{dns_unresolved})) * 100;
            $stats->{dns_effect_min} = $l->{dns_effect} if not $stats->{dns_effect_min} or $stats->{dns_effect_min} > $l->{dns_effect};
            $stats->{dns_effect_max} = $l->{dns_effect} if not $stats->{dns_effect_max} or $stats->{dns_effect_max} < $l->{dns_effect};
            $stats->{dns_effect_sum} += $l->{dns_effect};
            $stats->{dns_effect_cnt} += 1;
        }
        # Calculate DNS cache effectivity
        if ($l->{dns_cache_hit} + $l->{dns_cache_miss}) {
            $l->{dns_cache_effect} = ($l->{dns_cache_hit}/($l->{dns_cache_hit} + $l->{dns_cache_miss})) * 100;
            $stats->{dns_cache_effect_min} = $l->{dns_cache_effect} if not $stats->{dns_cache_effect_min} or $stats->{dns_cache_effect_min} > $l->{dns_cache_effect};
            $stats->{dns_cache_effect_max} = $l->{dns_cache_effect} if not $stats->{dns_cache_effect_max} or $stats->{dns_cache_effect_max} < $l->{dns_cache_effect};
            $stats->{dns_cache_effect_sum} += $l->{dns_cache_effect};
            $stats->{dns_cache_effect_cnt} += 1;
        }

        $i++;
        push(@{$stats->{graphs}->{'events_all'}->{data}->[0]}, $i);
        push(@{$stats->{graphs}->{'duration'}->{data}->[0]}, $i);
        push(@{$stats->{graphs}->{'speed_fetch'}->{data}->[0]}, $i);
        push(@{$stats->{graphs}->{'speed_aggr'}->{data}->[0]}, $i);
        push(@{$stats->{graphs}->{'dns_effect'}->{data}->[0]}, $i);
        push(@{$stats->{graphs}->{'dns_cache_effect'}->{data}->[0]}, $i);

        push(@{$stats->{graphs}->{'events_all'}->{data}->[1]}, $l->{events_all});
        push(@{$stats->{graphs}->{'duration'}->{data}->[1]}, $l->{duration});
        push(@{$stats->{graphs}->{'speed_fetch'}->{data}->[1]}, $l->{speed_fetch});
        push(@{$stats->{graphs}->{'speed_aggr'}->{data}->[1]}, $l->{speed_aggr});
        push(@{$stats->{graphs}->{'dns_effect'}->{data}->[1]}, $l->{dns_effect});
        push(@{$stats->{graphs}->{'dns_cache_effect'}->{data}->[1]}, $l->{dns_cache_effect});
    }

    # Calculate averages for various statistics
    foreach my $key (@stat_calc_keys, 'speed_fetch', 'speed_aggr', 'dns_effect', 'dns_cache_effect', 'reports_dest') {
        $stats->{$key.'_avg'} = $stats->{$key.'_sum'} / $stats->{$key.'_cnt'} if $stats->{$key.'_cnt'};
    }

    $self->_generate_charts_logs_reporting($stats, $adhoc);

    return $stats;
}

#-------------------------------------------------------------------------------

sub _ng_chart_pie_toplist($)
{
    my $self = shift;
    confess "Instance method not invoked on object instance" unless blessed($self);
    my ($chart_data, $threshold) = @_;

    $threshold -= 1;
    my $size = @$chart_data;
    my $rest = $size - $threshold;
    my $col  = Chart::Palette->color($threshold+1);
    if ($size > $threshold) {
        my $sum = 0;
        for (my $i = $threshold; $i < $size; $i++) {
            $sum += $chart_data->[$i]->{'value'};
        }

        $chart_data = [splice(@$chart_data, 0, $threshold)];
        push(@$chart_data, { 'label' => "-- Rest ($rest)", 'value' => $sum, 'color' => $col->[0] });
    }

    return $chart_data;
}

sub _ng_chart_pie($$)
{
    my $self = shift;
    confess "Instance method not invoked on object instance" unless blessed($self);
    my ($chart,) = @_;

    # Chart data
    my $d = $chart->{data};
    # Dataset
    my $ds = {};

    # Process the data in descending order
    my $col;
    my $i = 0;
    foreach my $key (sort {$d->{$b} <=> $d->{$a}} keys %$d) {
        $col = Chart::Palette->color(++$i);
        push(@{$ds->{'data'}}, { 'label' => $key, 'value' => $d->{$key}, 'color' => $col->[0] });

        $ds->{'stat_cnt'}++;
        $ds->{'stat_sum'} += $d->{$key};
        $ds->{'stat_min'}  = $d->{$key} if not $ds->{'stat_min'} or $ds->{'stat_min'} > $d->{$key};
        $ds->{'stat_max'}  = $d->{$key} if not $ds->{'stat_max'} or $ds->{'stat_max'} < $d->{$key};
    }
    # Calculate dataset average value
    $ds->{'stat_avg'} = $ds->{'stat_sum'} / $ds->{'stat_cnt'} if $ds->{'stat_cnt'};

    # Calculate dataset median
    my @values = sort { $a <=> $b } values %$d;
    my $size   = scalar @values;
    if ($size % 2) { $ds->{'stat_med'} = $values[((($size+1)/2)-1)]; }
    else           { $ds->{'stat_med'} = ($values[(($size/2)-1)] + $values[((($size+2)/2)-1)]) / 2; }

    # Perform dataset thresholding - create toplist
    $ds->{'data'} = $self->_ng_chart_pie_toplist($ds->{'data'}, $chart->{'toplist'});

    $chart->{'dataset'} = $ds;
    #delete $chart->{data};
    return $chart;
}

sub _ng_chart_bar($;$)
{
    my $self = shift;
    confess "Instance method not invoked on object instance" unless blessed($self);
    my ($chart, $replace) = @_;

    # Chart data
    my $d = $chart->{data};
    # Dataset
    my $ds = {};

    # Process the data in order
    my @values = ();
    my $col;
    my $i = 0;
    foreach my $data (@$d) {
        ++$i;
        $col = Chart::Palette->color_m($i);
        push(@values, $data->[1]);
        unless ($replace) {
            push(@{$ds->{'data'}}, { 'label' => $data->[0], 'value' => $data->[1], 'color' => $col->[0], 'id' => $data->[0] });
        }
        else {
            push(@{$ds->{'data'}}, { 'label' => $i, 'value' => $data->[1], 'color' => $col->[0], 'id' => $data->[0] });
        }

        $ds->{'stat_cnt'}++;
        $ds->{'stat_sum'} += $data->[1];
        $ds->{'stat_min'}  = $data->[1] if not $ds->{'stat_min'} or $ds->{'stat_min'} > $data->[1];
        $ds->{'stat_max'}  = $data->[1] if not $ds->{'stat_max'} or $ds->{'stat_max'} < $data->[1];
    }
    # Calculate dataset average value
    $ds->{'stat_avg'} = $ds->{'stat_sum'} / $ds->{'stat_cnt'} if $ds->{'stat_cnt'};

    # Calculate dataset median
    my $size   = scalar @values;
    if ($size % 2) { $ds->{'stat_med'} = $values[((($size+1)/2)-1)]; }
    else           { $ds->{'stat_med'} = ($values[(($size/2)-1)] + $values[((($size+2)/2)-1)]) / 2; }

    $chart->{'dataset'} = $ds;
    #delete $chart->{data};
    return $chart;
}

#-------------------------------------------------------------------------------

sub _ng_generate_charts_alertsstats($)
{
    my $self = shift;
    confess "Instance method not invoked on object instance" unless blessed($self);
    my ($stats,) = @_;

    # 'detectors', 'category_sets', 'categories', 'analyzers', 'detectorsws', 'ips'
    my @charts = ();
    push(@charts, {
            'chart_type' => 'pie',
            'dispv' => 'fmtn',
            'key' => 'detectors',
            'data' => _unescape_data($stats->{detectors}),
            'toplist' => TOP_THRESHOLD(),
            'label' => '# per detector',
            'title' => 'Number of events per detector',
        }) if $stats->{detectors};
    push(@charts, {
            'chart_type' => 'pie',
            'dispv' => 'fmtn',
            'key' => 'category_sets',
            'data' => _unescape_data($stats->{category_sets}),
            'toplist' => TOP_THRESHOLD(),
            'label' => '# per category set',
            'title' => 'Number of events per category set',
        }) if $stats->{category_sets};
    push(@charts, {
            'chart_type' => 'pie',
            'dispv' => 'fmtn',
            'key' => 'categories',
            'data' => _unescape_data($stats->{categories}),
            'toplist' => TOP_THRESHOLD(),
            'label' => '# per category',
            'title' => 'Number of events per category',
            'note' => 'Total number of events may be greater that the real number of events because CATEGORY is an attribute which can contain multiple values.',
        }) if $stats->{categories};
    push(@charts, {
            'chart_type' => 'pie',
            'dispv' => 'fmtn',
            'key' => 'analyzers',
            'data' => _unescape_data($stats->{analyzers}),
            'toplist' => TOP_THRESHOLD(),
            'label' => '# per analyzer',
            'title' => 'Number of events per analyzer',
            'note' => 'Total number of events may be greater that the real number of events because NODE/SW is an attribute which can contain multiple values.',
        }) if $stats->{analyzers};
    push(@charts, {
            'chart_type' => 'pie',
            'dispv' => 'fmtn',
            'key' => 'detectorsws',
            'data' => _unescape_data($stats->{detectorsws}),
            'toplist' => TOP_THRESHOLD(),
            'label' => '# per detector/SW',
            'title' => 'Number of events per detector/SW',
            'note' => 'Total number of events may be greater that the real number of events because NODE/SW is an attribute which can contain multiple values.',
        }) if $stats->{detectorsws};
    push(@charts, {
            'chart_type' => 'pie',
            'dispv' => 'fmtn',
            'key' => 'ips',
            'data' => _unescape_data($stats->{ips}),
            'toplist' => TOP_THRESHOLD(),
            'label' => '# per IP',
            'title' => 'Number of events per detector/SW',
            'note' => 'Total number of events may be different from the real number of events because SOURCE/IP is an attribute which can contain multiple values and also these statistics are calculated only from periodical toplists.',
        }) if $stats->{ips};

    foreach my $ch (@charts) {
        next unless $ch->{'data'};
        if (ref $ch->{'data'} eq 'ARRAY') {
            next unless scalar @{$ch->{'data'}};
        }
        else {
            next unless scalar keys %{$ch->{'data'}};
        }
        push(@{$stats->{charts}}, $self->_ng_chart_pie($ch)) if $ch->{'chart_type'} eq 'pie';
        push(@{$stats->{charts}}, $self->_ng_chart_bar($ch)) if $ch->{'chart_type'} eq 'bar';
    }
}

sub _ng_generate_charts_accesslogs($)
{
    my $self = shift;
    confess "Instance method not invoked on object instance" unless blessed($self);
    my ($stats,) = @_;

    my @charts = (
            {
                'chart_type' => 'pie',
                'dispv' => 'fmtn',
                'key' => 'users',
                'data' => $stats->{users},
                'toplist' => TOP_THRESHOLD(),
                'label' => '# per user',
                'title' => 'Number of hits per user',
            },
            {
                'chart_type' => 'pie',
                'dispv' => 'fmtn',
                'key' => 'paths',
                'data' => $stats->{paths},
                'toplist' => TOP_THRESHOLD(),
                'label' => '# per path',
                'title' => 'Number of hits per path',
            },
        );

    foreach my $ch (@charts) {
        next unless $ch->{'data'};
        if (ref $ch->{'data'} eq 'ARRAY') {
            next unless scalar @{$ch->{'data'}};
        }
        else {
            next unless scalar keys %{$ch->{'data'}};
        }
        push(@{$stats->{charts}}, $self->_ng_chart_pie($ch)) if $ch->{'chart_type'} eq 'pie';
        push(@{$stats->{charts}}, $self->_ng_chart_bar($ch)) if $ch->{'chart_type'} eq 'bar';
    }
}

sub _ng_generate_charts_report($;$)
{
    my $self = shift;
    confess "Instance method not invoked on object instance" unless blessed($self);
    my ($stats, $single) = @_;

    # 'abuses', 'size_abuses', 'mails', 'analyzers', 'categories', 'ips'
    my @charts = (
            {
                'chart_type' => 'pie',
                'dispv' => 'fmtn',
                'key' => 'analyzers',
                'data' => _unescape_data($stats->{analyzers}),
                'toplist' => TOP_THRESHOLD(),
                'label' => '# per analyzer',
                'title' => 'Number of events per analyzer',
                'note' => 'Total number of events may be greater that the real number of events because NODE/SW is an attribute which can contain multiple values.',
            },
            {
                'chart_type' => 'pie',
                'dispv' => 'fmtn',
                'key' => 'categories',
                'data' => _unescape_data($stats->{categories}),
                'toplist' => TOP_THRESHOLD(),
                'label' => '# per category',
                'title' => 'Number of events per category',
                'note' => 'Total number of events may be greater that the real number of events because CATEGORY is an attribute which can contain multiple values.',
            },
            {
                'chart_type' => 'pie',
                'dispv' => 'fmtn',
                'key' => 'ips',
                'data' => _unescape_data($stats->{ips}),
                'toplist' => TOP_THRESHOLD(),
                'label' => '# per IP',
                'title' => 'Number of events per IP',
                'note' => 'Total number of events may be greater that the real number of events because SOURCE/IP is an attribute which can contain multiple values.',
            },
        );

    foreach my $ch (@charts) {
        next unless $ch->{'data'};
        if (ref $ch->{'data'} eq 'ARRAY') {
            next unless scalar @{$ch->{'data'}};
        }
        else {
            next unless scalar keys %{$ch->{'data'}};
        }
        push(@{$stats->{charts}}, $self->_ng_chart_pie($ch))    if $ch->{'chart_type'} eq 'pie';
        push(@{$stats->{charts}}, $self->_ng_chart_bar($ch, 1)) if $ch->{'chart_type'} eq 'bar';
    }
}

sub _ng_generate_charts_reports($;$$)
{
    my $self = shift;
    confess "Instance method not invoked on object instance" unless blessed($self);
    my ($stats, $single, $briefs) = @_;

    # 'abuses', 'size_abuses', 'mails'

    #Log::Loger->notice(__PACKAGE__, "[MSM] STATS: -- ".Dumper($stats));

    my @charts = (
            {
                'chart_type' => 'pie',
                'dispv' => 'fmtn',
                'key' => 'detectors',
                'data' => _unescape_data($stats->{detectors}),
                'toplist' => TOP_THRESHOLD(),
                'label' => '# per detector',
                'title' => 'Number of events per detector',
            },
            {
                'chart_type' => 'pie',
                'dispv' => 'fmtn',
                'key' => 'category_sets',
                'data' => _unescape_data($stats->{category_sets}),
                'toplist' => TOP_THRESHOLD(),
                'label' => '# per category set',
                'title' => 'Number of events per category set',
            },
            {
                'chart_type' => 'pie',
                'dispv' => 'fmtn',
                'key' => 'analyzers',
                'data' => _unescape_data($stats->{analyzers}),
                'toplist' => TOP_THRESHOLD(),
                'label' => '# per analyzer',
                'title' => 'Number of events per analyzer',
                'note' => 'Total number of events may be greater that the real number of events because NODE/SW is an attribute which can contain multiple values.',
            },
            {
                'chart_type' => 'pie',
                'dispv' => 'fmtn',
                'key' => 'categories',
                'data' => _unescape_data($stats->{categories}),
                'toplist' => TOP_THRESHOLD(),
                'label' => '# per category',
                'title' => 'Number of events per category',
                'links' => [['alerts_categ','alerts']],
                'note' => 'Total number of events may be greater that the real number of events because CATEGORY is an attribute which can contain multiple values.',
            },
            {
                'chart_type' => 'pie',
                'dispv' => 'fmtn',
                'key' => 'ips',
                'data' => _unescape_data($stats->{ips}),
                'toplist' => TOP_THRESHOLD(),
                'label' => '# per IP',
                'title' => 'Number of events per IP',
                'links' => [['alerts_ip','alerts'],['reports','reports']],
                'note' => 'Total number of events may be greater that the real number of events because SOURCE/IP is an attribute which can contain multiple values.',
            },
        );
        unless($briefs) {
            unshift(@charts,
                {
                    'chart_type' => 'bar',
                    'dispv' => 'fmtn',
                    'key' => 'counts_all',
                    'data' => $stats->{counts_all},
                    'label' => '# per report',
                    'title' => 'Number of events per report',
                    'replaced' => 1,
                    'links' => [['report','view']],
                }
            );
        }
        unless($single) {
            push(@charts,
                {
                    'chart_type' => 'pie',
                    'dispv' => 'fmtn',
                    'key' => 'abuses',
                    'data' => _unescape_data($stats->{abuses}),
                    'toplist' => TOP_THRESHOLD(),
                    'label' => '@ per abuse',
                    'title' => 'Number of emails per abuse group',
                    'links' => [['group','view'],['reports','reports']],
                },
                {
                    'chart_type' => 'pie',
                    'dispv' => 'fmtn',
                    'key' => 'mails',
                    'data' => _unescape_data($stats->{mails}),
                    'toplist' => TOP_THRESHOLD(),
                    'label' => '@ per address',
                    'title' => 'Number of emails per email address',
                },
                {
                    'chart_type' => 'bar',
                    'dispv' => 'fmtn',
                    'key' => 'delays',
                    'data' => $stats->{delays},
                    'label' => 'Delays',
                    'title' => 'Reporting delays per report',
                    'replaced' => 1,
                    'links' => [['report','view']],
                },
            );
        }

    #Log::Loger->notice(__PACKAGE__, "[MSM] CHARTS: -- ".Dumper(@charts));

    foreach my $ch (@charts) {
        next unless $ch->{'data'};
        if (ref $ch->{'data'} eq 'ARRAY') {
            next unless scalar @{$ch->{'data'}};
        }
        else {
            next unless scalar keys %{$ch->{'data'}};
        }
        push(@{$stats->{charts}}, $self->_ng_chart_pie($ch))    if $ch->{'chart_type'} eq 'pie';
        push(@{$stats->{charts}}, $self->_ng_chart_bar($ch, 1)) if $ch->{'chart_type'} eq 'bar';
    }
}

#-------------------------------------------------------------------------------

=item ng_evaluate_alert_stats($;$) [PUBLIC]

  Usage     : my $stats = $reportstats->ng_evaluate_alert_stats($messages, $stats);
  Purpose   : Evaluate statistics for list of alerts
  Arguments : ARRAY REFERENCE $messages - list of messages [MANDATORY]
  Returns   : HASH REFERENCE $stats
  Throws    : Dies, if not invoked on instance

=cut

sub ng_evaluate_alert_stats($;$)
{
    my $self = shift;
    confess "Instance method not invoked on object instance" unless blessed($self);
    my ($objects, $adhoc) = @_;

    my $stats = {};
    foreach my $obj (@$objects) {
        foreach my $t (ST_OVERALL(), ST_EXTERNAL(), ST_INTERNAL()) {
            next unless exists $obj->{$t};

            if ($t eq ST_OVERALL()) {
                $stats->{cnt_all} += $obj->{count};
                $stats->{ts_from} = $obj->{ts_from} if not $stats->{ts_from} or $stats->{ts_from} > $obj->{ts_from};
                $stats->{ts_to}   = $obj->{ts_to}   if not $stats->{ts_to}   or $stats->{ts_to}   < $obj->{ts_to};
            }

            $stats->{$t}->{cnt_all}    += $obj->{count};
            $stats->{$t}->{cnt_alerts} += $obj->{$t}->{cnt_alerts};

            $stats->{$t}->{ts_from} = $obj->{ts_from} if not $stats->{$t}->{ts_from} or $stats->{$t}->{ts_from} > $obj->{ts_from};
            $stats->{$t}->{ts_to}   = $obj->{ts_to}   if not $stats->{$t}->{ts_to}   or $stats->{$t}->{ts_to}   < $obj->{ts_to};

            foreach my $s ('analyzers', 'categories', 'category_sets', 'detectors', 'detectorsws', 'ips') {
                next unless exists $obj->{$t}->{'cnt_'.$s};
                $stats->{$t}->{'cnt_'.$s} += $obj->{$t}->{'cnt_'.$s};
                foreach my $d (keys %{$obj->{$t}->{$s}}) {
                    $stats->{$t}->{$s}->{$d} += $obj->{$t}->{$s}->{$d};
                }
            }
        }
    }

    # Statistics postprocessing
    my $s;
    foreach my $sg (ST_OVERALL(), ST_EXTERNAL(), ST_INTERNAL()) {
        next unless exists $stats->{$sg};
        $s = $stats->{$sg}; # Only for shorter names

        $self->_ng_generate_charts_alertsstats($s);
    }

    # Generate graphs
    $self->_generate_charts_alert_stats($stats, $adhoc);

    return $stats;
}

=item ng_evaluate_accesslogs($;$) [PUBLIC]

  Usage     : my $stats = $accessstats->ng_evaluate_accesslogs($logs);
  Purpose   : Evaluate statistics for accesslogs
  Arguments : ARRAY REFERENCE $logs - list of accesslogs [MANDATORY]
  Returns   : HASH REFERENCE $stats
  Throws    : Dies, if not invoked on instance

=cut

sub ng_evaluate_accesslogs($;$)
{
    my $self = shift;
    confess "Instance method not invoked on object instance" unless blessed($self);
    my ($logs, $adhoc) = @_;
    die "Accesslogs must be given as argument\n" unless $logs;

    # Storage for calculated statistics
    my $stats = {};
    return $stats unless $logs and ref $logs eq 'ARRAY' and scalar @$logs;

    foreach my $l (@$logs) {
        $stats->{cnt_all}++;

        # Calculate the time period
        $stats->{ts_from} = $l->{ts_from} if not $stats->{ts_from} or $stats->{ts_from} > $l->{ts_from};
        $stats->{ts_to}   = $l->{ts_to}   if not $stats->{ts_to}   or $stats->{ts_to}   < $l->{ts_to};

        $stats->{users}->{$l->{user}}++;
        $stats->{paths}->{'/'.$l->{path}}++;
    }

    $self->_ng_generate_charts_accesslogs($stats);

    return $stats;
}

=item ng_evaluate_report($) [PUBLIC]

  Usage     : my $stats = $reportstats->ng_evaluate_report($report);
  Purpose   : Evaluate statistics for single report
  Arguments : ARRAY REFERENCE $messages - list of messages [MANDATORY]
  Returns   : HASH REFERENCE $stats
  Throws    : Dies, if not invoked on instance

=cut

sub ng_evaluate_report($)
{
    my $self = shift;
    confess "Instance method not invoked on object instance" unless blessed($self);
    my ($report,) = @_;
    die "Report must be given as argument\n" unless $report;

    my $stats = {};
    $stats->{'id'} = $report->{_id};

    # Prepare data for additional graphs
    foreach my $key ('analyzers', 'categories', 'ips') {
        $stats->{$key} = $report->{$key};
    }

    # Generate charts
    $self->_ng_generate_charts_report($stats);

    # Return the resulting statistics
    return $stats;
}

=item ng_evaluate_reports($) [PUBLIC]

  Usage     : my $stats = $reportstats->ng_evaluate_reports($reports);
  Purpose   : Evaluate statistics for report list
  Arguments : ARRAY REFERENCE $messages - list of messages [MANDATORY]
  Returns   : HASH REFERENCE $stats
  Throws    : Dies, if not invoked on instance

=cut

sub ng_evaluate_reports($;$$)
{
    my $self = shift;
    confess "Instance method not invoked on object instance" unless blessed($self);
    my ($report_list, $single, $briefs) = @_;
    die "Report list must be given as argument\n" unless $report_list;

    # Storage for calculated statistics
    my $stats = {ST_INTERNAL() => {}, ST_EXTERNAL() => {}, ST_OVERALL() => {}};

    # List of various keys for which to calculate statistics
    my @stat_calc_keys  = ('cnt_all', 'cnt_detectors', 'cnt_category_sets', 'cnt_categories', 'cnt_analyzers', 'cnt_ips');

    my ($s, $d);
    foreach my $obj (@$report_list) {
        # List of stat types to be updated
        my @stl = (ST_OVERALL());
        # Determine the other type of stat type, which will be updated
        push(@stl, (($obj->{abuse} =~ /UNKNOWN/) ? ST_EXTERNAL() : ST_INTERNAL()));

        $stats->{cnt_reports}++;

        $stats->{cnt_all} += $obj->{cnt_all} || 0;
        $stats->{cnt_mails}++ if $obj->{mail_to};
        $stats->{ts_from} = $obj->{ts_from} if not $stats->{ts_from} or $stats->{ts_from} > $obj->{ts_from};
        $stats->{ts_to}   = $obj->{ts_to}   if not $stats->{ts_to}   or $stats->{ts_to}   < $obj->{ts_to};

        foreach my $st (@stl) {
            $s = $stats->{$st}; # Only for shorter names

            # Update information about involved abuse groups
            $s->{cnt_abuses}++;
            $s->{abuses}->{$obj->{abuse}}++;

            # Update information about involved email destination
            if ($obj->{mail_to}) {
                $s->{cnt_mails}++;
                $s->{mails}->{$obj->{mail_to}}++;
            }

            # Update information about report types
            $s->{cnt_summary}++ if $obj->{type} eq 'summary';
            $s->{cnt_extra}++   if $obj->{type} eq 'extra';

            # Update information about report severities
            $s->{cnt_sev_low}++      if $obj->{severity} eq 'low';
            $s->{cnt_sev_medium}++   if $obj->{severity} eq 'medium';
            $s->{cnt_sev_high}++     if $obj->{severity} eq 'high';
            $s->{cnt_sev_critical}++ if $obj->{severity} eq 'critical';

            # Following calculations should be done only for summary reports, extra
            # reports would be redundant
            next unless $obj->{type} eq 'summary';

            # Calculate the total report count
            $s->{count}++;
            # Calculate the total event count
            $s->{cnt_all} += $obj->{cnt_all};

            # Calculate the event magnitude for each abuse group
            $s->{size_abuses}->{$obj->{abuse}} += $obj->{cnt_all};

            # Calculate the time period
            $s->{ts_from} = $obj->{ts_from} if not $s->{ts_from} or $s->{ts_from} > $obj->{ts_from};
            $s->{ts_to}   = $obj->{ts_to}   if not $s->{ts_to}   or $s->{ts_to}   < $obj->{ts_to};

            # Calculate statistics for report delays
            $d = $obj->{ts} - $obj->{ts_to};
            $s->{delay_min}  = $d if not $s->{delay_min} or $s->{delay_min} > $d;
            $s->{delay_max}  = $d if not $s->{delay_max} or $s->{delay_max} < $d;
            $s->{delay_sum} += $d;
            $s->{delay_cnt} += 1;

            # Calculate batch of other report statistics
            foreach my $key (@stat_calc_keys) {
                $s->{$key.'_min'}  = $obj->{$key} if not $s->{$key.'_min'} or $s->{$key.'_min'} > $obj->{$key};
                $s->{$key.'_max'}  = $obj->{$key} if not $s->{$key.'_max'} or $s->{$key.'_max'} < $obj->{$key};
                $s->{$key.'_sum'} += $obj->{$key};
                $s->{$key.'_cnt'} += 1;
            }

            # Mark the analyzers and categories contained within the report
            foreach my $c (keys %{$obj->{detectors}}) {
                $s->{detectors}->{$c} += $obj->{detectors}->{$c};
            }
            foreach my $c (keys %{$obj->{category_sets}}) {
                $s->{category_sets}->{$c} += $obj->{category_sets}->{$c};
            }
            foreach my $c (keys %{$obj->{categories}}) {
                $s->{categories}->{$c} += $obj->{categories}->{$c};
            }
            foreach my $a (keys %{$obj->{analyzers}}) {
                $s->{analyzers}->{$a} += $obj->{analyzers}->{$a};
            }
            foreach my $i (keys %{$obj->{ips}}) {
                $s->{ips}->{$i} += $obj->{ips}->{$i};
            }

            push(@{$s->{'counts_all'}}, [$obj->{_id}, $obj->{cnt_all}]);
            push(@{$s->{'delays'}},     [$obj->{_id}, $d]);
        }
    }

    # Generate charts
    foreach my $sg (ST_OVERALL(), ST_EXTERNAL(), ST_INTERNAL()) {
        $s = $stats->{$sg}; # Only for shorter names

        $self->_ng_generate_charts_reports($s, $single, $briefs);
    }

    # Return the resulting statistics
    return $stats;
}

#-------------------------------------------------------------------------------

=item evaluate_duration($$;$$) [PUBLIC]

  Usage     : my $stats = $stats->evaluate_duration($type, $runlogs, $stats);
  Purpose   : Evaluate statistics for list of alerts
  Arguments : ARRAY REFERENCE $messages - list of messages [MANDATORY]
  Returns   : HASH REFERENCE $stats
  Throws    : Dies, if not invoked on instance

=cut

sub evaluate_duration($$;$$)
{
    my $self = shift;
    confess "Instance method not invoked on object instance" unless blessed($self);
    my ($type, $runlogs, $stats, $adhoc) = @_;

    $runlogs = [$runlogs] unless ref $runlogs eq 'ARRAY';
    $stats   = {}         unless $stats and ref $stats eq 'HASH';

    my $i = 0;
    foreach my $rl (@$runlogs) {
        $stats->{ts_from} = $rl->{ts} if not $stats->{ts_from} or $stats->{ts_from} > $rl->{ts};
        $stats->{ts_to}   = $rl->{ts} if not $stats->{ts_to}   or $stats->{ts_to}   < $rl->{ts};

        unshift(@{$stats->{data}->[0]}, $rl->{ts});
        unshift(@{$stats->{data}->[1]}, $rl->{duration});

        $stats->{min}  = $rl->{duration} if not $stats->{min} or $stats->{min} > $rl->{duration};
        $stats->{max}  = $rl->{duration} if not $stats->{max} or $stats->{max} < $rl->{duration};
        $stats->{sum} += $rl->{duration};
        $stats->{cnt}++;
    }
    $stats->{avg} = $stats->{sum} / $stats->{cnt} if $stats->{cnt};

    $self->_generate_charts_duration_sl($type, $stats, $adhoc);

    #print STDERR "STATS:".Dumper($stats);

    return $stats;
}

#-------------------------------------------------------------------------------

sub charter()
{
    my $self = shift;
    confess "Instance method not invoked on object instance" unless blessed($self);
    return $self->{CHARTER};
}

#-------------------------------------------------------------------------------
# Private interface
#-------------------------------------------------------------------------------

# _init [PROTECTED]
#
#  Usage     : return $self->_init(@_);
#  Purpose   : Initialize the new Mentat::Stats::Data instance
#  Arguments : Unknown see the description in appropriate subclass
#  Returns   : Mentat::Stats::Data $self reference
#  Throws    : Dies, if not invoked on instance or given invalid arguments
#  Comments  : This method must never be called directly, it is supposed to be overloaded by subclasses.

sub _init
{
    my $self = shift;
    confess "Instance method not invoked on object instance" unless blessed($self);
    ### Initializing Mentat::Stats::Data instance...
    my %config = @_;

    $self->{CHARTER} = Mentat::Stats::Charter->new(%config);

    return $self;
}

#-------------------------------------------------------------------------------

sub _chart_toplist_threshold($)
{
    my $self = shift;
    confess "Instance method not invoked on object instance" unless blessed($self);
    my ($chart_data,) = @_;

    my $size = @{$chart_data->[0]};
    my $rest = $size - TOP_THRESHOLD();
    if ($size > TOP_THRESHOLD()) {
        $chart_data->[0] = [splice(@{$chart_data->[0]}, 0, TOP_THRESHOLD())];
        push(@{$chart_data->[0]}, "-- Rest ($rest)");
        my $sum = 0;
        foreach my $v (splice(@{$chart_data->[1]}, TOP_THRESHOLD())) {
            $sum += $v;
        }
        $chart_data->[1] = [splice(@{$chart_data->[1]}, 0, TOP_THRESHOLD())];
        push(@{$chart_data->[1]}, $sum);
    }

    return $chart_data;
}

sub _chart_generate_prefix($$$$)
{
    my $self = shift;
    confess "Instance method not invoked on object instance" unless blessed($self);
    my ($type, $from, $to, $rnd) = @_;

    if ($rnd) {
        return '_'.$type.'_'.$rnd.'_';
    }
    else {
        my $sf = Value::Convertor->unixs_to_ldatestr($from);
        my $st = Value::Convertor->unixs_to_ldatestr($to);

        my $l = ''.$type.'_'.$st.'_'.$sf.'_';
        $l =~ s/[-: ]//g;
        return $l;
    }
}

#-------------------------------------------------------------------------------

sub _generate_charts_alert_stats($$)
{
    my $self = shift;
    confess "Instance method not invoked on object instance" unless blessed($self);
    my ($stats, $adhoc) = @_;

    my ($s, $l, @graphs, $type);
    my $rnd = ($adhoc) ? Mentat::Core::uid(undef, 8) : '';
    foreach my $sg (ST_OVERALL(), ST_EXTERNAL(), ST_INTERNAL()) {
        next unless exists $stats->{$sg};
        $s = $stats->{$sg}; # Only for shorter names

        if ($s->{graphs}) {
            die "Missing input data to generate the 'categories' graph\n"  unless $s->{graphs}->{categories};
            die "Missing input data to generate the 'analyzers' graph\n"   unless $s->{graphs}->{analyzers};

            $l = $self->_chart_generate_prefix('alertsstats', $s->{ts_from}, $s->{ts_to}, $rnd);

            @graphs = (
                # Full sized charts
                ['pie', $s->{graphs}->{'detectors'},      $l.'detectors'."_$sg",     'fn'],
                ['pie', $s->{graphs}->{'category_sets'},  $l.'category_sets'."_$sg", 'fn'],
                ['pie', $s->{graphs}->{'categories'},     $l.'categories'."_$sg",    'fn'],
                ['pie', $s->{graphs}->{'analyzers'},      $l.'analyzers'."_$sg",     'fn'],
                ['pie', $s->{graphs}->{'ips'},            $l.'ips'."_$sg",           'fn'],
            );
            foreach my $g (@graphs) {
                $type = shift @$g;
                $self->charter->chart($type, @$g);
                $self->charter->chart('sls_'.$type, @$g);
            }
        }
    }
}

sub _generate_charts_report($$)
{
    my $self = shift;
    confess "Instance method not invoked on object instance" unless blessed($self);
    my ($stats, $adhoc) = @_;

    my ($l, @graphs, $type);
    if ($stats->{charts}) {
        $l = 'report_'.$stats->{id};

        @graphs = ();
        push (@graphs, ['pie', $stats->{charts}->{'categories'},  $l.'categories',  'fn']) if $stats->{charts}->{categories};
        push (@graphs, ['pie', $stats->{charts}->{'analyzers'},   $l.'analyzers',   'fn']) if $stats->{charts}->{analyzers};
        push (@graphs, ['pie', $stats->{charts}->{'ips'},         $l.'ips',         'fn']) if $stats->{charts}->{ips};

        foreach my $g (@graphs) {
            $type = shift @$g;
            $self->charter->chart($type, @$g);
            $self->charter->chart('sls_'.$type, @$g);
        }
    }
}

sub _generate_charts_reports($$)
{
    my $self = shift;
    confess "Instance method not invoked on object instance" unless blessed($self);
    my ($stats, $adhoc) = @_;

    my ($s, $l, @graphs, $type);
    my $rnd = ($adhoc) ? Mentat::Core::uid(undef, 8) : '';
    foreach my $sg (ST_OVERALL(), ST_EXTERNAL(), ST_INTERNAL()) {
        $s = $stats->{$sg};
        if ($s->{graphs}) {
            die "Missing input data to generate the 'cnt_all' graph\n"     unless $s->{graphs}->{cnt_all};
            die "Missing input data to generate the 'delay' graph\n"       unless $s->{graphs}->{delay};
            die "Missing input data to generate the 'abuses' graph\n"      unless $s->{graphs}->{abuses};
            die "Missing input data to generate the 'size_abuses' graph\n" unless $s->{graphs}->{size_abuses};
            die "Missing input data to generate the 'analyzers' graph\n"   unless $s->{graphs}->{analyzers};
            die "Missing input data to generate the 'categories' graph\n"  unless $s->{graphs}->{categories};

            $l   = $self->_chart_generate_prefix('reports', $s->{ts_from}, $s->{ts_to}, $rnd);

            @graphs = (
                ['bar', $s->{graphs}->{'cnt_all'},     $l.'cnt_all'."_$sg",     'fn',  1],
                ['bar', $s->{graphs}->{'delay'},       $l.'delay'."_$sg",       'dur', 1],
                ['pie', $s->{graphs}->{'abuses'},      $l.'abuses'."_$sg",      'fn'],
                ['pie', $s->{graphs}->{'mails'},       $l.'mails'."_$sg",       'fn'],
                ['pie', $s->{graphs}->{'size_abuses'}, $l.'size_abuses'."_$sg", 'fn'],
                ['pie', $s->{graphs}->{'analyzers'},   $l.'analyzers'."_$sg",   'fn'],
                ['pie', $s->{graphs}->{'categories'},  $l.'categories'."_$sg",  'fn'],
                ['pie', $s->{graphs}->{'ips'},         $l.'ips'."_$sg",         'fn'],
            );
            foreach my $g (@graphs) {
                $type = shift @$g;
                $self->charter->chart($type, @$g);
                $self->charter->chart('sls_'.$type, @$g);
            }
        }
    }
}

sub _generate_charts_logs_reporting($$)
{
    my $self = shift;
    confess "Instance method not invoked on object instance" unless blessed($self);
    my ($stats, $adhoc) = @_;
    die "Missing input data to generate the 'events_all' graph\n"       unless $stats->{graphs}->{events_all};
    die "Missing input data to generate the 'duration' graph\n"         unless $stats->{graphs}->{duration};
    die "Missing input data to generate the 'speed_fetch' graph\n"      unless $stats->{graphs}->{speed_fetch};
    die "Missing input data to generate the 'speed_aggr' graph\n"       unless $stats->{graphs}->{speed_aggr};
    die "Missing input data to generate the 'dns_effect' graph\n"       unless $stats->{graphs}->{dns_effect};
    die "Missing input data to generate the 'dns_cache_effect' graph\n" unless $stats->{graphs}->{dns_cache_effect};

    my $rnd = ($adhoc) ? Mentat::Core::uid(undef, 8) : '';
    my $l   = $self->_chart_generate_prefix('reports', $stats->{ts_from}, $stats->{ts_to}, $rnd);

    my @graphs = (
            ['bar', $stats->{graphs}->{'events_all'},        $l.'_events_all',       'fn'],
            ['bar', $stats->{graphs}->{'duration'},          $l.'_duration',         'dur'],
            ['bar', $stats->{graphs}->{'speed_fetch'},       $l.'_speed_fetch',      'fn'],
            ['bar', $stats->{graphs}->{'speed_aggr'},        $l.'_speed_aggr',       'fn'],
            ['bar', $stats->{graphs}->{'dns_effect'},        $l.'_dns_effect',       'fn'],
            ['bar', $stats->{graphs}->{'dns_cache_effect'},  $l.'_dns_cache_effect', 'fn'],
        );

    my $type;
    foreach my $g (@graphs) {
        $type = shift @$g;
        $self->charter->chart($type, @$g);
        $self->charter->chart('sls_'.$type, @$g);
    }
}

sub _generate_charts_duration_sl($$$)
{
    my $self = shift;
    confess "Instance method not invoked on object instance" unless blessed($self);
    my ($type, $stats, $adhoc) = @_;

    my $rnd = ($adhoc) ? Mentat::Core::uid(undef, 8) : '';
    my $l   = $self->_chart_generate_prefix($type, $stats->{ts_from}, $stats->{ts_to}, $rnd);

    my @graphs = (
            ['bar', $stats, $l.'_duration', 'dur'],
        );

    my $gtype;
    foreach my $g (@graphs) {
        $gtype = shift @$g;
        #$self->charter->chart($gtype, @$g);
        $self->charter->chart('sls_'.$gtype, @$g);
    }
}

=pod

=back

=cut

1;
